from ..systems import DiscreteSys
from ..tools import nd_rand_init

import numpy as np


class LogisticMap(DiscreteSys):
    def __init__(self, r=4, n_points=100):
        def map_func(x):
            return r * x * (1 - x)

        def rand_init():
            return nd_rand_init((0, 1))

        super().__init__(dim=1, map_func=map_func, init_func=rand_init, n_points=n_points)



from ..tools import correlation_sum, delay_coordinates, reference_rule
from ..tools.io_conversion import _make_array
from ..tools.nd_utils import nd_function, windowed_function
from ..tools.correlation_sum import _fast_count_row
import numpy as np


@nd_function
@windowed_function
def sample_entropy(x, radius=None, dim=2, norm_p=float('inf'), axis=0):
    _x = _make_array(x)

    Xm = delay_coordinates(_x, dim=dim, lag=1, return_array=True, axis=axis)
    Xm_1 = delay_coordinates(_x, dim=dim + 1, lag=1, return_array=True, axis=axis)

    if radius is None:
        radius = reference_rule(Xm, dim=dim, norm_p=float('inf'))

    B = correlation_sum(Xm, r=radius, norm_p=norm_p)
    A = correlation_sum(Xm_1, r=radius, norm_p=norm_p)

    return - np.log(A / B)


@nd_function
@windowed_function
def approximate_entropy(x, radius=None, dim=2, norm_p=float('inf'), axis=0):
    _x = _make_array(x)

    Xm = delay_coordinates(_x, dim=dim, lag=1, return_array=True, axis=axis)
    Xm_1 = delay_coordinates(_x, dim=dim + 1, lag=1, return_array=True, axis=axis)

    if radius is None:
        radius = reference_rule(Xm, dim=dim, norm_p=float('inf'))

    Bi = np.apply_along_axis(_fast_count_row, 1, Xm, Xm, radius, norm_p).astype(np.float64)
    Ai = np.apply_along_axis(_fast_count_row, 1, Xm_1, Xm_1, radius, norm_p).astype(np.float64)

    logAi = np.log(Ai)
    logAi = logAi[logAi == logAi]

    logBi = np.log(Bi)
    logBi = logBi[logBi == logBi]

    return np.mean(logBi) - np.mean(logAi)


def ks_entropy():
    pass



import numpy as np
from pyunicorn.timeseries import RecurrencePlot
from ..tools import weighted_white_vertline_histogram, weighted_vertline_histogram, weighted_diagline_histogram


class TimeResolvedRecurrencePlot(RecurrencePlot):
    def __init__(self, time_series, metric="supremum", normalize=False,
                 missing_values=False, sparse_rqa=False, silence_level=0, default_scale=None,
                 **kwds):
        super().__init__(time_series, metric, normalize, missing_values, sparse_rqa, silence_level, **kwds)

        self._time_resolved_diagline_dist = None
        self._time_resolved_diagline_dist_scale = 0
        self._time_resolved_diagline_dist_order = 0

        self._time_resolved_vertline_dist = None
        self._time_resolved_vertline_dist_scale = 0
        self._time_resolved_vertline_dist_order = 0

        self._time_resolved_white_vertline_dist = None
        self._time_resolved_white_vertline_dist_scale = 0
        self._time_resolved_white_vertline_dist_order = 0

        if default_scale is None:
            default_scale = int(0.05 * self.N)

        self._default_scale = default_scale

    def time_resolved_diagline_dist(self, scale=None, order=2):
        if scale is not None:
            assert (scale > 0)
        if (
                self._time_resolved_diagline_dist_scale == scale or scale is None) and self._time_resolved_diagline_dist_order == order:
            return self._time_resolved_diagline_dist
        else:
            if scale is None:
                scale = self._default_scale
            recmat = self.recurrence_matrix()
            self._time_resolved_diagline_dist = weighted_diagline_histogram(recmat, scale, order)
            self._time_resolved_diagline_dist_scale = scale
            self._time_resolved_diagline_dist_order = order

            return self._time_resolved_diagline_dist

    def time_resolved_vertline_dist(self, scale=None, order=2):
        if scale is not None:
            assert (scale > 0)
        if (
                self._time_resolved_vertline_dist_scale == scale or scale is None) and self._time_resolved_vertline_dist_order == order:
            return self._time_resolved_vertline_dist
        else:
            if scale is None:
                scale = self._default_scale
            recmat = self.recurrence_matrix()
            self._time_resolved_vertline_dist = weighted_vertline_histogram(recmat, scale, order)
            self._time_resolved_vertline_dist_scale = scale
            self._time_resolved_vertline_dist_order = order
            return self._time_resolved_vertline_dist

    def time_resolved_white_vertline_dist(self, scale=None, order=2):
        if scale is not None:
            assert (scale > 0)
        if (
                self._time_resolved_white_vertline_dist_scale == scale or scale is None) and self._time_resolved_white_vertline_dist_order == order:
            return self._time_resolved_white_vertline_dist
        else:
            if scale is None:
                scale = self._default_scale
            recmat = self.recurrence_matrix()
            self._time_resolved_white_vertline_dist = weighted_white_vertline_histogram(recmat, scale, order)
            self._time_resolved_white_vertline_dist_scale = scale
            self._time_resolved_white_vertline_dist_order = order
            return self._time_resolved_white_vertline_dist

    def time_resolved_determinism(self, l_min=2, scale=None, order=2):
        diagline = self.time_resolved_diagline_dist(scale, order)

        if l_min > diagline.shape[1]:
            return np.zeros((self.N,))

        partial_sum = (np.arange(l_min, diagline.shape[1])[None, :] * diagline[:, l_min:]).sum(axis=1)
        full_sum = (np.arange(diagline.shape[1])[None] * diagline).sum(axis=1).astype(float)

        return partial_sum / (full_sum + self._epsilon)

    def time_resolved_max_diaglength(self, scale=None, order=2):
        diagline = self.time_resolved_diagline_dist(scale, order)
        return diagline.argmax(axis=1)

    def time_resolved_average_diaglength(self, l_min=2, scale=None, order=2):
        diagline = self.time_resolved_diagline_dist(scale, order)

        if l_min > diagline.shape[1]:
            return np.zeros((self.N,))

        partial_sum = (np.arange(l_min, diagline.shape[1])[None, :] * diagline[:, l_min:]).sum(axis=1)
        full_sum = diagline[:, l_min:].sum(axis=1).astype(float)
        return partial_sum / full_sum + self._epsilon

    def time_resolved_diag_entropy(self, l_min=2, scale=None, order=2):
        diagline = self.time_resolved_diagline_dist(scale, order)

        if l_min > diagline.shape[1]:
            # TODO: check this
            return np.zeros((self.N,))

        diagline = diagline[:, l_min:]
        diagnorm = diagline / (diagline.sum(axis=1, keepdims=True).astype(float) + self._epsilon)
        valid_entries = diagline > 0
        entr = np.zeros_like(diagline)
        entr[valid_entries] = -(diagnorm[valid_entries] * np.log(diagnorm[valid_entries]))

        return entr.sum(axis=1)

    def time_resolved_laminarity(self, v_min=2, scale=None, order=2):
        vertline = self.time_resolved_vertline_dist(scale, order)

        if v_min > vertline.shape[1]:
            return np.zeros((self.N,))

        partial_sum = (np.arange(v_min, vertline.shape[1])[None, :] * vertline[:, v_min:]).sum(axis=1)
        full_sum = (np.arange(vertline.shape[1])[None] * vertline).sum(axis=1).astype(float)

        return partial_sum / (full_sum + self._epsilon)

    def time_resolved_max_vertlength(self, scale=None, order=2):
        vertline = self.time_resolved_vertline_dist(scale, order)
        return vertline.argmax(axis=1)

    def time_resolved_average_vertlength(self, v_min=2, scale=None, order=2):
        vertline = self.time_resolved_vertline_dist(scale, order)

        if v_min > vertline.shape[1]:
            return np.zeros((self.N,))

        partial_sum = (np.arange(v_min, vertline.shape[1])[None, :] * vertline[:, v_min:]).sum(axis=1)
        full_sum = vertline[:, v_min:].sum(axis=1).astype(float)
        return partial_sum / full_sum + self._epsilon

    def time_resolved_trapping_time(self, v_min=2, scale=None, order=2):
        return self.time_resolved_average_vertlength(v_min, scale, order)

    def time_resolved_vert_entropy(self, v_min=2, scale=None, order=2):
        vertline = self.time_resolved_vertline_dist(scale, order)

        if v_min > vertline.shape[1]:
            # TODO: check this
            return np.zeros((self.N,))

        vertline = vertline[:, v_min:]
        vertnorm = vertline / (vertline.sum(axis=1, keepdims=True).astype(float) + self._epsilon)
        valid_entries = vertline > 0
        entr = np.zeros_like(vertline)
        entr[valid_entries] = -(vertnorm[valid_entries] * np.log(vertnorm[valid_entries]))

        return entr.sum(axis=1)

    def time_resolved_max_white_vertlength(self, scale=None, order=2):
        white_vertline = self.time_resolved_white_vertline_dist(scale, order)
        return white_vertline.argmax(axis=1)

    def time_resolved_average_white_vertlength(self, w_min=2, scale=None, order=2):
        white_vertline = self.time_resolved_white_vertline_dist(scale, order)

        if w_min > white_vertline.shape[1]:
            return np.zeros((self.N,))

        partial_sum = (np.arange(w_min, white_vertline.shape[1])[None, :] * white_vertline[:, w_min:]).sum(axis=1)
        full_sum = white_vertline[:, w_min:].sum(axis=1)
        return partial_sum / (full_sum.astype(float) + self._epsilon)

    def time_resolved_mean_recurrence_time(self, w_min=2, scale=None, order=2):
        return self.time_resolved_average_white_vertlength(w_min, scale, order)

    def time_resolved_white_vert_entropy(self, w_min=2, scale=None, order=2):
        white_vertline = self.time_resolved_white_vertline_dist(scale, order)

        if w_min > white_vertline.shape[1]:
            # TODO: check this
            return np.zeros((self.N,))

        white_vertline = white_vertline[:, w_min:]
        white_vertnorm = white_vertline / (white_vertline.sum(axis=1, keepdims=True).astype(float) + self._epsilon)
        valid_entries = white_vertline > 0
        entr = np.zeros_like(white_vertline)
        entr[valid_entries] = -(white_vertnorm[valid_entries] * np.log(white_vertnorm[valid_entries]))

        return entr.sum(axis=1)


from ..tools import reference_rule, correlation_sum
from ..tools.math_utils import _lstsqr_design_matrix
from ..tools.io_conversion import _make_array
from ..tools.nd_utils import nd_function, windowed_function

import numpy as np


@nd_function
@windowed_function
def correlation_dimension(x, radius=None, radius_range=(0.5, 1), use_relative_range=True,
                          n_radius=10, log_base=np.e, norm_p=float('inf'), debug_plot=False):

    """ Computes the correlation dimension from a timeseries
    """

    _x = _make_array(x)

    if radius is None:
        radius = reference_rule(_x, norm_p=norm_p)

    if log_base == 10:
        log = np.log10
    elif log_base == np.e:
        log = np.log
    else:
        log = lambda x: np.log(_x) / np.log(log_base)

    if use_relative_range:
        absolute_range = list(b * radius for b in radius_range)
    else:
        absolute_range = list(radius_range)
    absolute_range = [log(r) for r in absolute_range]

    radius_values = np.logspace(absolute_range[0], absolute_range[1], n_radius, base=log_base)
    corr_sums = np.vectorize(lambda r: correlation_sum(_x, r, norm_p=norm_p))(radius_values)

    radius_values = log(radius_values[corr_sums != 0])
    corr_sums = log(corr_sums[corr_sums != 0])

    poly = np.linalg.lstsq(_lstsqr_design_matrix(radius_values), corr_sums, rcond=None)[0]
    corr_dim = poly[0]

    if debug_plot:
        import plotly.graph_objects as go

        fig = go.Figure()
        fig.add_trace(go.Scatter(x=radius_values, y=corr_sums, name="log(C(r)) vs log(r)"))
        fig.add_trace(go.Scatter(x=radius_values, y=poly[1] + radius_values * poly[0],
                                 name="%.2f log(r) + %.2f"%(poly[0], poly[1]), line=dict(dash="dash"),
                                 marker=dict(symbol="x-thin")))
        fig.show()

    return corr_dim




from .math_utils import n_ball_volume, n_sphere_area, nd_rand_init, make_uniform_kernel, mutual_information, lagged_mi, lstsqr, sigmoid

from .embedding import select_embedding_lag, delay_coordinates

from .io_conversion import from_array, to_array, from_pandas_series, to_pandas_series, from_data_frame, to_data_frame, \
    from_mne_raw


from .corr_utils import *
from .plot_utils import plot_rp, Scale4Latex
from .correlation_sum import correlation_sum
from .recurrence_plot_utils import localized_diagline_histogram, localized_vertline_histogram, \
    localized_white_vertline_histogram, weighted_vertline_histogram, weighted_diagline_histogram, \
    weighted_white_vertline_histogram



from .io_conversion import _make_array
import numpy as np
from typing import Union


def _fast_count_row(x, traj, r, norm_p):
    """
    :param x: (dim, )
    :param traj: (n_points, dim)
    :param norm_p: int, float('inf')
    :return:
    """
    return np.count_nonzero(np.linalg.norm(traj - x[np.newaxis, :], ord=norm_p, axis=1) <= r)


def _corr_sum_numpy_backend(x: np.ndarray, r: float, norm_p: float) -> float:
    return np.sum(
        np.apply_along_axis(_fast_count_row, 1, x, x, r, norm_p)).astype(np.float64) / (x.shape[0]**2)


def _corr_sum_pyunicorn_backend(x: np.ndarray, r: float, norm_p: Union[float, str]) -> float:
    import pyunicorn.timeseries as puts
    from .io_conversion import _to_pyunicorm_metric

    return puts.RecurrencePlot(
        x, threshold=r, metric=_to_pyunicorm_metric(norm_p), silence_level=3
    ).recurrence_rate()


try:
    import pyunicorn
    _corr_sum_backend = _corr_sum_pyunicorn_backend
except ImportError:
    print("Cannot import pyunicorn backend. Custom Numpy backend will be use instead (could be"
          "slower.")
    _corr_sum_backend = _corr_sum_numpy_backend


def correlation_sum(x, r=None, norm_p=float('inf'), allow_equals=False) -> float:
    _x = _make_array(x)
    csum = _corr_sum_backend(x, r, norm_p)
    if not allow_equals:
        N = _x.shape[0]
        csum = (csum * N - 1)/(N-1)
    return csum


from ..tools import reference_rule

from scipy.special import erf
from scipy.signal import correlate
from scipy.stats import multivariate_normal

import numpy as np


def autocov(x, axis=-1):
    if len(x.shape) > 1:
        return np.apply_along_axis(autocov, axis, x)
    else:
        return correlate(x, x, mode='full', method='auto')[-x.size:] / x.size


class ParametricRQA:
    """
    Computes the asymptotic quantities of recurrence quantification analysis from timeseries autocorrelation, as
    described in Ramdani, Sofiane, et al. "Parametric recurrence quantification analysis of autoregressive processes
    for pattern recognition in multichannel electroencephalographic data." Pattern Recognition 109 (2021): 107572.
    """
    def __init__(self, x, axis=-1, threshold=None, lmin=2, vmin=2):
        self.x = x
        self.axis = axis
        if threshold is None:
            threshold = reference_rule(x)
        self.threshold = threshold
        self.lmin = lmin
        self.vmin = vmin
        self._internals = dict(diag_probs=dict(), vert_probs=dict())

    def _get_autocov(self):
        try:
            return self._internals['autocov']
        except KeyError:
            self._internals['autocov'] = autocov(self.x, self.axis)
            return self._internals['autocov']

    def _get_diag_prob(self, k):
        try:
            return self._internals['diag_probs'][k]
        except KeyError:
            self._internals['diag_probs'][k] = self._compute_diag_prob(k)
            return self._internals['diag_probs'][k]

    def _get_vert_prob(self, k):
        try:
            return self._internals['vert_probs'][k]
        except KeyError:
            self._internals['vert_probs'][k] = self._compute_vert_prob(k)
            return self._internals['vert_probs'][k]

    def recurrence_rate(self):
        rr = erf(self.threshold / (2 * self.x.std(ddof=1, axis=self.axis)))
        if isinstance(rr, np.ndarray) and rr.size == 1:
            rr = rr[0]
        return rr

    def determinism(self, lmin=None):
        if lmin is None:
            lmin = self.lmin
        Pk = self._get_diag_prob(lmin)
        Pk_1 = self._get_diag_prob(lmin + 1)
        P1 = self._get_diag_prob(1)
        return (lmin * Pk - (lmin - 1) * Pk_1) / P1

    def avg_diagline(self, lmin=None):
        if lmin is None:
            lmin = self.lmin
        Pk = self._get_diag_prob(lmin)
        Pk_1 = self._get_diag_prob(lmin + 1)
        return (lmin * Pk - (lmin - 1) * Pk_1) / (Pk - Pk_1)

    def laminarity(self, vmin=None):
        if vmin is None:
            vmin = self.vmin
        Tk = self._get_vert_prob(vmin)
        Tk_1 = self._get_vert_prob(vmin + 1)
        T1 = self._get_vert_prob(1)
        return (vmin * Tk - (vmin - 1) * Tk_1) / T1

    def avg_vertline(self, vmin=None):
        if vmin is None:
            vmin = self.vmin
        Tk = self._get_vert_prob(vmin)
        Tk_1 = self._get_vert_prob(vmin + 1)
        return (vmin * Tk - (vmin - 1) * Tk_1) / (Tk - Tk_1)

    def rqa_summary(self, lmin=None, vmin=None):
        return dict(
            rec=self.recurrence_rate(),
            det=self.determinism(lmin),
            lavg=self.avg_diagline(lmin),
            lam=self.laminarity(vmin),
            vavg=self.avg_vertline(vmin)
        )

    def _compute_diag_prob(self, k):
        axis = self.axis
        autocov = self._get_autocov()
        epsilon = self.threshold

        if axis != -1:
            autocov = np.swapaxes(autocov, axis, -1)

        r = np.repeat(np.arange(k)[:, None], k, axis=1)
        s = r.T
        idxs = r - s
        q = autocov.shape[-1] - k

        c1 = np.take(autocov, np.abs(idxs), axis=-1)
        c2 = np.take(autocov, np.abs(idxs + q), axis=-1)
        c3 = np.take(autocov, np.abs(idxs - q), axis=-1)

        cov = 2 * c1[..., :, :] - c2 - c3

        eps = np.repeat(epsilon, k)

        shape = cov.shape[:-2]
        cov = cov.reshape((-1, k, k))
        cov = (cov + np.swapaxes(cov, 1, 2)) / 2

        prob = []
        for _cov in cov:
            try:
                res = multivariate_normal.cdf(eps, cov=_cov)
            except ValueError:
                res = np.nan
            except np.linalg.LinAlgError:
                res = np.nan
            prob.append(res)

        if len(prob) > 1:
            prob = np.asarray(prob).reshape(shape)

            if axis != -1:
                prob = np.swapaxes(prob, -1, axis)
        else:
            prob = prob[0]

        return prob

    def _compute_vert_prob(self, k):
        axis = self.axis
        autocov = self._get_autocov()
        epsilon = self.threshold

        if axis != -1:
            autocov = np.swapaxes(autocov, axis, -1)

        r = np.repeat(np.arange(k)[:, None], k, axis=1)
        s = r.T
        idxs = r - s

        # todo: decrement q iterativelly until cov is SPD
        q = autocov.shape[-1] - k

        c1 = autocov[..., 0]
        # Note: in the two following lines, s - 1 and r - 1 from the paper is replaced
        # with s and r as they range from 0 to k-1 instead of 1 to k. This does not change
        # anything for the difference r - s
        c2 = np.take(autocov, np.abs(q - s), axis=-1)
        c3 = np.take(autocov, np.abs(r - q), axis=-1)
        c4 = np.take(autocov, np.abs(idxs), axis=-1)

        cov = c1 - c2 - c3 + c4

        eps = np.repeat(epsilon, k)

        shape = cov.shape[:-2]
        cov = cov.reshape((-1, k, k))
        cov = (cov + np.swapaxes(cov, 1, 2)) / 2

        prob = []
        for _cov in cov:
            try:
                res = multivariate_normal.cdf(eps, cov=_cov)
            except ValueError:
                res = np.nan
            except np.linalg.LinAlgError:
                res = np.nan
            prob.append(res)

        if len(prob) > 1:
            prob = np.asarray(prob).reshape(shape)

            if axis != -1:
                prob = np.swapaxes(prob, -1, axis)
        else:
            prob = prob[0]

        return prob



from .correlation_dimension import correlation_dimension
from .entropies import sample_entropy, approximate_entropy
from .ParametricRQA import ParametricRQA
from .WindowedRecurrencePlot import WindowedRecurrencePlot
from .WindowedParametricRQA import WindowedParametricRQA
from .TimeResolvedRecurrencePlot import TimeResolvedRecurrencePlot



from ..tools import lagged_mi
from ..systems import AbstractBaseSys, SysWrapper

import numpy as np
import pandas as pd
import plotly.graph_objs as go

from typing import Union


def delay_coordinates(
        ts: np.ndarray,
        dim: int,
        lag: Union[int, str] = "auto",
        axis: int = 0,
        return_array: bool = False,
        *args, **kwargs
    ) -> Union[np.ndarray, SysWrapper]:
    if len(ts.shape) == 1:
        ts = ts[:, None]
        axis = 0

    if lag == "auto":
        lag = select_embedding_lag(ts, axis=axis, *args, **kwargs)
    if lag * dim > ts.shape[0]:
        raise ValueError

    data = np.vstack([ts[i * lag:(i - dim) * lag, axis] for i  in range(dim)]).T


    if return_array:
        return data
    else:
        return SysWrapper(data)


def select_embedding_lag(
        data: Union[np.ndarray, AbstractBaseSys, pd.Series],
        axis: int = 0,
        lag_range: Union[int, tuple] = (1, 500),
        method: str = "mi",
        criterion: Union[float, str] = 'firstmin',
        interactive: bool = False,
        plot: bool = False
    ) -> int:

    assert(method in ['acf', 'mi'])
    if isinstance(criterion, str):
        assert(criterion == 'firstmin')

    if interactive:
        raise NotImplementedError()

    ts = None
    if method == 'acf': # convert to pandas
        if isinstance(data, np.ndarray):
            ts = pd.Series(data[:,axis])
        elif isinstance(data, AbstractBaseSys):
            from ..tools.io_conversion import to_pandas_series
            ts = to_pandas_series(data, axis=axis)
        elif isinstance(data, pd.Series):
            ts = data.copy()
    else: # convert to np
        if isinstance(data, np.ndarray):
            if len(data.shape) > 1:
                data = data[:, axis]
            ts = np.asarray(data)
        elif isinstance(data, AbstractBaseSys):
            ts = data.states[:, axis]
        elif isinstance(data, pd.Series):
            ts = data.to_numpy(copy=True)


    lag_range = (lag_range[0], min(lag_range[1], ts.size) - 1)
    dcurve = None
    ref = None
    lag = -1

    if method == "acf":
        ref = ts.autocorr(0)
        dcurve = [ts.autocorr(lag) for lag in range(*lag_range)]
    elif method == "mi":
        ref = lagged_mi(ts, 0)
        dcurve = [lagged_mi(ts, lag) for lag in range(*lag_range)]

    if isinstance(criterion, str):
        _dcurve = [dcurve[i+1] - dcurve[i] for i in range(len(dcurve) - 1)]

        for i in range(len(_dcurve) - 1):
            if _dcurve[i] < 0 and _dcurve[i + 1] > 0:
                lag = i if abs(dcurve[i]) < abs(dcurve[i + 1]) else i + 1
                break
    else:
        for i in range(len(dcurve) - 1):
            if dcurve[i] > ref * criterion and dcurve[i + 1] <= ref * criterion:
                lag = i if abs(dcurve[i]) < abs(dcurve[i + 1]) else i + 1
                break

    if plot:
        fig = go.Figure()
        fig.add_trace(go.Scatter(y=dcurve, mode="lines"))
        fig.add_shape(x0=lag, x1=lag, y0=np.min(dcurve), y1=np.max(dcurve), type="line", line_color="indianred")
        fig.show()

    return lag










import numpy as np


from ..systems import ContinuousSys
from ..tools import nd_rand_init


class Lorenz(ContinuousSys):
    def __init__(self, sigma=10., beta=8. / 3, rho=28., n_points=10000, t_min=0, t_max=100):
        def ode(X, t):
            _x, _y, _z = X
            return np.asarray([
                sigma * (_y - _x),
                _x * (rho - _z) - _y,
                _x * _y - beta * _z
            ])

        def rand_init():
            return nd_rand_init([-20, 20], [-30, 30], [10, 40])

        super().__init__(dim=3, map_func=ode, init_func=rand_init, n_points=n_points, t_min=t_min, t_max=t_max)



from .ParametricRQA import ParametricRQA
from ..tools.io_conversion import _make_array, _from_pyunicorn_metric
from ..tools.corr_utils import reference_rule
from tqdm import tqdm
import numpy as np
import functools
from joblib import Parallel, delayed

#todo select threshold per window or overall
class WindowedParametricRQA:
    def __init__(self, x, window_size, n_overlap=None, time_axis=None, state_axis=None, n_jobs=None, **kwargs):
        """

        """
        x = _make_array(x)
        self._time_axis = time_axis
        self._state_axis = state_axis
        self._n_jobs = n_jobs

        self._is_ndim = False
        if len(x.shape) > 2 or (len(x.shape) == 2 and state_axis == None):
            assert (time_axis is not None)
            self._is_ndim = True

        if time_axis is not None:

            self._original_shape = x.shape

            if state_axis is None:
                self._has_state_dim = False
                x = np.swapaxes(x, time_axis, -1)
                x = np.expand_dims(x, -1)
            else:
                self._has_state_dim = True
                x = np.moveaxis(x, (time_axis, state_axis), (-2, -1))

            self._unrelated_shape = x.shape[:-2]
            self._related_shape = x.shape[-2:]

        else:
            assert (state_axis is None)
            x = np.expand_dims(x, -1)

            self._unrelated_shape = tuple()
            self._related_shape = x.shape[-2:]

        x = x.reshape((-1, *self._related_shape))

        self._window_size = window_size
        if n_overlap is None:
            n_overlap = window_size // 2
        self._n_overlap = n_overlap
        self._step_size = window_size - n_overlap
        self._n_windows = x.shape[-2] // self._step_size

        self._time_series = [
            _x[i * self._step_size:min(x.shape[-2] + 1, i * self._step_size + self._window_size), :]
            for _x in x for i in range(self._n_windows)
        ]

        if not any(
                k in ('threshold', )
                for k in kwargs.keys()):

            norm_p = _from_pyunicorn_metric(kwargs.pop('metric')) if 'metric' in kwargs else _from_pyunicorn_metric(
                'supremum')
            # radii = [reference_rule(ts, norm_p=norm_p) for ts in self._time_series]

            self._rps = [ParametricRQA(ts, threshold=reference_rule(ts, norm_p=norm_p), axis=0, **kwargs)
                         for ts in tqdm(self._time_series)]

        else:
            self._rps = [ParametricRQA(ts, axis=0, **kwargs)
                         for ts in tqdm(self._time_series)
                         ]

        for name in dir(ParametricRQA):
            method = getattr(ParametricRQA, name)
            if callable(method) and name[0] != '_':
                self._bind_rps_method(name, method)

    def _bind_rps_method(self, name, method):
        @functools.wraps(method)
        def _wrapped_func(*args, **kwargs):
            if self._n_jobs is None:
                results = []
                for rp in tqdm(self._rps):
                    results.append(method(rp, *args, **kwargs))
            else:
                results = Parallel(self._n_jobs)(delayed(method)(rp, *args, **kwargs)
                    for rp in self._rps)
            results = np.asarray(results)
            if self._is_ndim:
                if self._has_state_dim:
                    results = results.reshape((*self._unrelated_shape, -1, self._related_shape[1]))
                    results = np.moveaxis(results, (-1, -2), (self._state_axis, self._time_axis))
                else:
                    results = results.reshape((*self._unrelated_shape, -1))
                    results = np.swapaxes(results, -1, self._time_axis)

            if isinstance(results.flat[0], dict): 
                unpack = lambda key: np.vectorize(lambda _: _[key])
                results = {k: unpack(k)(results) for k in results.flat[0].keys()}

            return results

        setattr(self, method.__name__, _wrapped_func)


from ..systems import AbstractBaseSys
from ..tools import reference_rule

import plotly.graph_objs as go
import plotly.express as px
from plotly.subplots import make_subplots



import numpy as np

from typing import Union

black_white_colorscale = [[0, "rgb(255, 255, 255)"],[1, "rgb(0, 0, 0)"]]


def plot_rp(sys: Union[AbstractBaseSys, None] = None,
            rp: Union[np.ndarray, None] = None,
            sig: Union[np.ndarray, None] = None,
            time_vec: Union[np.ndarray, None] = None,
            width: int = 1080,
            rp_height_ratio: float = 0.8,
            vspacing_ratio: float = 0.1,
            axis: int = 0
        ) -> go.Figure:

    _time_vec = None
    if time_vec is None:
        _time_vec = sys.time_vec if sys is not None else None
    print(_time_vec)

    _ts = None
    if sys is None:
        assert(sig is not None)
        _ts = sig
    else:
        _ts = sys.states

    _rp = None
    if rp is None:
        import pyunicorn.timeseries as puts
        _rp = puts.RecurrencePlot(_ts, threshold=reference_rule(_ts, norm_p='inf')).R
    elif isinstance(rp, np.ndarray):
        _rp = rp
    # elif isinstance(rp, puts.RecurrencePlot):
    #     _rp = rp.R

    fig = make_subplots(rows=2, row_heights=[rp_height_ratio, 1-rp_height_ratio-vspacing_ratio], vertical_spacing=vspacing_ratio)

    fig.add_trace(go.Heatmap(z=_rp, colorscale=black_white_colorscale,  showscale=False, yaxis="y"), row=1, col=1)
    fig.add_trace(go.Scatter(x=_time_vec, y=_ts[:, axis], line=dict(color="rgb(0,0,0)")), row=2, col=1)

    fig.update_yaxes(col=1, row=1, autorange='reversed')
    fig.update_layout(width=width, height=int(width / rp_height_ratio))

    return fig


class Scale4Latex:
    DEFAULTS = {
        'img_width': 1920,
        'img_height': 1080,
        'font_size': 10,
        'title_font_size': 11,
        'font_family': 'Computer Modern',
        'line_width': 1,
        'marker_size': 4,
    }

    def __init__(self, output_img_width=1920, latex_img_width=624, **kwargs):
        self._scale = output_img_width / latex_img_width
        self._img_width = output_img_width
        self._latex_img_width = latex_img_width
        for k, v in self.__class__.DEFAULTS.items():
            value = kwargs[k] if k in kwargs.keys() else v
            if isinstance(value, int) or isinstance(value, float):
                self.__dict__.update({k: self._scale_object(value)})
            else:
                self.__dict__.update({k:value})

    def _scale_object(self, o):
        return int(self._scale * o)

    @property
    def img_width(self):
        return self._img_width

    @property
    def latex_img_width(self):
        return self._latex_img_width

    def scale_font(self, font_size):
        return self._scale_object(font_size)


from ..systems import ContinuousSys
from ..tools import nd_rand_init

import numpy as np


class VanDerPol(ContinuousSys):
    def __init__(self, mu=2., n_points=1000, t_min=0, t_max=100):
        def ode(X, t):
            x, y = X
            return np.asarray([
                y,
                (1. - x**2) * mu * y - x
            ])

        def rand_init():
            return nd_rand_init((-2, 2), (-2, 2))

        super().__init__(dim=2, map_func=ode, init_func=rand_init, n_points=n_points, t_min=t_min, t_max=t_max)




from ..systems import AbstractBaseSys, SysWrapper
import pandas as pd
import numpy as np
import mne
from typing import Union


def to_data_frame(sys: AbstractBaseSys, columns=None) -> pd.DataFrame:
    return pd.DataFrame(data=sys.states, index=sys.time_vec, columns=columns)


def from_data_frame(df: pd.DataFrame) -> SysWrapper:
    return SysWrapper(df, dim=len(df.columns), n_points=len(df.index), time_vec=np.asarray(df.index))


def to_pandas_series(sys: AbstractBaseSys, axis: int = 0) -> pd.Series:
    if sys.dim > 1:
        return pd.Series(data=sys.states[:, axis], index=sys.time_vec)
    else:
        return pd.Series(data=sys.states, index=sys.time_vec)


def from_pandas_series(ts: pd.Series) -> SysWrapper:
    return SysWrapper(ts, dim=1, n_points=len(ts.index), time_vec=np.asarray(ts.index))


def to_array(sys: AbstractBaseSys) -> np.ndarray:
    return sys.states


def from_array(a: np.ndarray, *args, **kwargs):
    return SysWrapper(a, *args, **kwargs)


def from_mne_raw(raw: mne.io.BaseRaw) -> SysWrapper:
    return SysWrapper(raw.get_data().T, time_vec=np.asarray(raw.times))


def _make_array(ts_or_sys: Union[np.ndarray, AbstractBaseSys]) -> np.ndarray:
    return ts_or_sys if isinstance(ts_or_sys, np.ndarray) else to_array(ts_or_sys)


_pyunicorn_np_map = {'manhattan':1, 'euclidean':2, 'supremum':float('inf')}
_np_pyunicorn_map = {v:k for k, v in _pyunicorn_np_map.items()}

_pyunicorn_np_map.update({v:v for v in _pyunicorn_np_map.values()})
_np_pyunicorn_map.update({v:v for v in _np_pyunicorn_map.values()})


def _to_pyunicorm_metric(norm_p):
    return _np_pyunicorn_map[norm_p]


def _from_pyunicorn_metric(metric):
    return _pyunicorn_np_map[metric]



from ..systems import ContinuousSys
from ..tools import nd_rand_init

import numpy as np


class HindmarshRose(ContinuousSys):
    def __init__(self, I=3.25, a=1, b=3, c=1, d=5, r=1e-3, s=4, x_r=-(1 + np.sqrt(5))/2., n_points=5000, t_min=0, t_max=30):

        def ode(X, t):
            x, y, z = X
            return np.asarray([
                y - a * x**3 + b * x**2 - z + I,
                c - d * x**2 - y,
                r * (s * (x - x_r) - z)
            ])

        def rand_init():
            return nd_rand_init((-1, 2), (-1, 1), (-1, 1))

        super().__init__(dim=3, map_func=ode, init_func=rand_init, n_points=n_points, t_min=t_min, t_max=t_max)




from ..systems import DiscreteSys
from ..tools import nd_rand_init

import numpy as np


class HenonMap(DiscreteSys):
    def __init__(self, a=1.4, b=0.3, n_points=100):
        def map_func(X):
            x, y = X
            return np.asarray([1. - a * x**2 + y, b * x])

        def rand_init():
            return nd_rand_init((-0.5, 0.5), (-0.5, 0.5))

        super().__init__(dim=2, map_func=map_func, init_func=rand_init, n_points=n_points)



import numpy as np
from ..tools import n_ball_volume, n_sphere_area, delay_coordinates, lstsqr
from ..tools.nd_utils import nd_function
from .math_utils import _lstsqr_design_matrix
from scipy.special import gamma
from nolds.measures import poly_fit
from tqdm import tqdm
from typing import Union

import plotly.graph_objs as go


def _fast_count_row_1d(x, traj, r, norm_p):
    return np.count_nonzero(np.linalg.norm(traj - x[np.newaxis, :], ord=norm_p, axis=1) <= r)


def _fast_count_row(x, traj, r, norm_p):
    """
    :param x: (dim, )
    :param traj: (n_points, dim)
    :param norm_p: int, float('inf')
    :return:
    """
    return np.count_nonzero(np.linalg.norm(traj - x[np.newaxis, :], ord=norm_p, axis=1) <= r)


def _fast_count_traj(x, r, norm_p):
    """
    :param x: (n_points, dim)
    :param norm_p: int, float('inf')
    :return:
    """
    if x.shape[1] > 1:
        return np.sum(np.apply_along_axis(_fast_count_row, 1, x, x, r, norm_p))
    elif x.shape[1] == 1:
        return np.sum(np.apply_along_axis(_fast_count_row_1d, 1, x, x, r, norm_p))


def corr_sum(traj, r, norm_p=1, allow_equals=False):
    if allow_equals:
        return _fast_count_traj(traj, r, norm_p).astype(np.float64) / traj.shape[0] ** 2
    else:
        return (_fast_count_traj(traj, r, norm_p).astype(np.float64) - traj.shape[0]) /\
               (traj.shape[0] * (traj.shape[0] - 1))


def dsty_est(x, samples, r, norm_p=1):
    if len(samples.shape) == 1:
        samples = samples[:, np.newaxis]

    dim = samples.shape[1]

    # x is (k,)
    if len(x.shape) == 1:
        # if k = d, then x is a point in phase space
        if x.shape[0] == dim:
            x = x[np.newaxis, :]
        # x is a collection of sample points in 1d phase space
        elif dim == 1:
            x = x[:, np.newaxis]
    elif len(x.shape) == 2:
        assert(x.shape[1] == samples.shape[1])

    return np.apply_along_axis(_fast_count_row, 1, x, samples, r, norm_p).astype(np.float64) / (n_ball_volume(dim, norm_p) * r**dim * samples.shape[0])


def reference_rule_alpha(p: Union[float, int], d: int):
    if d == 1:
        alpha = (12. * np.sqrt(np.pi)) ** (1./5)
    else:
        alpha = (
            (4. * (2. * np.sqrt(np.pi))**d * (3. * gamma(1+(d+2.)/p) * gamma(1+1./p))**2)
            / ((d + 2.) * n_ball_volume(d, p) * (gamma(1+3./p) * gamma(1.+d/p)) ** 2)
        ) ** (1./(d+4))
    return alpha


@nd_function
def reference_rule(x: np.ndarray, dim: Union[int, str] = 'auto', norm_p: Union[int, float, str] = 2) -> float:
    n = x.shape[0]
    if dim == 'auto':
        d = 1
        if len(x.shape) == 2:
            d = x.shape[1]

    elif isinstance(dim, int):
        d = dim
    else:
        raise ValueError('dim must be "auto" or int')

    # print(d)
    std = np.sqrt(x.var(axis=0, ddof=1).mean())
    from scipy import stats
    iqr = stats.iqr(x)
    scale = min(std, iqr/1.34)

    gamma_n = n ** (-1/(d+4.))

    if norm_p in ['manhattan', 'euclidean', 'supremum']:
        norm_p = ["manhattan", "euclidean"].index(norm_p) + 1 if norm_p != "supremum" else float("inf")

    alpha_p_d = reference_rule_alpha(norm_p, d)
    return gamma_n * alpha_p_d * scale

# def rule_of_thumb(x: np.ndarray, norm_p=2, version: str = 'normal') -> float:
#     n = x.shape[0]
#     d = 1
#     if len(x.shape) == 2:
#         d = x.shape[1]
#
#     if norm_p in ['manhattan', 'euclidean', 'supremum']:
#         norm_p = ["manhattan", "euclidean"].index(norm_p) + 1 if norm_p != "supremum" else float("inf")
#
#     std = np.sqrt(x.var(axis=0, ddof=1).mean())
#
#
#     # version 1
#     if version == 'normal':
#         return std * ((9. * n_ball_volume(d, norm_p) * (2 * np.sqrt(np.pi)) ** d)/ ((d + 2) * n)) ** (1/(d+4))
#     elif version == 'scott':
#         return std * 3.5 * n ** (-1 / (d + 2))
#     # version 2
#     # return std * (((d + 2)**2 * (2*np.sqrt(np.pi))**d) / (n * n_ball_volume(d, norm_p) * (1/2. * d + 1/4. * d**2))) ** (1/(d+4))


def grassberger_proccacia(x: np.ndarray, rvals=None, rmin=None, rmax=None, omit_plateau=True, norm_p=2, method='lstsqr',
                          hack_filter_rvals=None,  nr=20, plot=False, fig=None, show=True, full_output=False,
                          log_base=10, remove_tail=True, verbose=False):
    """
    Estimates the correlation dimension using the Grassberger-Proccacia algorithm. The code is greatly inspired by
    nolds: https://github.com/CSchoel/nolds/blob/master/nolds/measures.py and makes use of nolds version of poly_fit
    :param x: time-series (n_points, dim) or (n_points, )
    :param rvals: the threshold values to use
    :return: the correlation dimension
    """
    if log_base == 10:
        log = np.log10
    elif log_base == np.exp(1):
        log = np.log
    else:
        log = lambda x: np.log(x)/np.log(log_base)

    if rvals is None:
        if rmin is not None and rmax is not None:
            rvals = np.logspace(rmin, rmax, nr, base=log_base)
        else:
            rvals = np.logspace(- log(x.shape[0]) + log(x.std()) - 2, 5 + log(x.std()), nr, base=log_base)
    # print(rvals)
    csums = np.asarray([corr_sum(x, r, norm_p=norm_p)
                        for r in (tqdm(rvals, desc="Computing correlation sums") if verbose else rvals)])
    # print(csums)
    orig_log_csums =log(csums[csums > 0])
    orig_log_rvals = log(rvals[csums > 0])

    log_csums = np.asarray(orig_log_csums)
    log_rvals = np.asarray(orig_log_rvals)


    if remove_tail:
        filter = log_csums > -log(x.shape[0])
        log_csums = log_csums[filter]
        log_rvals = log_rvals[filter]

    if hack_filter_rvals is not None:
        log_rvals = log_rvals[hack_filter_rvals]
        log_csums = log_csums[hack_filter_rvals]

    if omit_plateau:
        filter = np.zeros_like(log_rvals, dtype=np.bool)
        for i in range(log_rvals.size - 1):
            delta = log_csums[i+1] - log_csums[i]
            if delta > 0:
                filter[i] = True

        log_rvals = log_rvals[filter]
        log_csums = log_csums[filter]

    # poly = poly_fit(log_rvals, log_csums, degree=1)
    if len(log_rvals) == 0:
        raise ValueError('bad estim')

    if method == 'lstsqr':
        poly = lstsqr(_lstsqr_design_matrix(log_rvals), log_csums)
    else:
        poly = np.polyfit(log_rvals, log_csums, deg=1)

    if plot:
        if fig is None:
            fig = go.Figure()
        fig.add_trace(go.Scatter(x=orig_log_rvals, y=orig_log_csums, name="log(C(r)) vs log(r)"))
        fig.add_trace(go.Scatter(x=orig_log_rvals, y=poly[1] + orig_log_rvals * poly[0],
                                 name="%.2f log(r) + %.2f"%(poly[0], poly[1]), line=dict(dash="dash"),
                                 marker=dict(symbol="x-thin")))

        if show:
            fig.show()

    if full_output:
        return [orig_log_rvals, orig_log_csums, poly]
    else:
        return poly[0]


def approximate_d2(x: object, r_opt: object = None, meaningfull_range: object = (0.5, 1.), n_evals: object = 10, base: object = 10.,
                   norm_p: float = float('inf'),
                   method: object = 'fit',
                   full_output: object = False, output_curve = False) -> float:
    if r_opt is None:
        r_opt = reference_rule(x, norm_p=norm_p)
    if base ==10:
        log = np.log10
    elif base == np.e:
        log = np.log
    else:
        log = lambda x: np.log(x) / np.log(base)

    rvals = np.logspace(log(r_opt * meaningfull_range[0]), log(r_opt*meaningfull_range[1]), n_evals, base=base)
    _csum =lambda r : corr_sum(x, r, norm_p=norm_p)
    csums = np.vectorize(_csum)(rvals)
    csums = log(csums)
    csums = csums[csums == csums]
    rvals = log(rvals)

    if method == 'tangent':
        slopes = []
        for i in range(1, len(rvals)):
            rsup = rvals[i]
            for j in range(i):
                rinf = rvals[j]
                slopes.append((csums[i] - csums[j])/(rsup - rinf))

        return np.mean(slopes)
    elif method == 'fit':
        poly = lstsqr(_lstsqr_design_matrix(rvals), csums)
        # poly = poly_fit(rvals, csums, degree=1)
        if full_output:
            if output_curve:
                return poly[0], poly[1],  log(r_opt), rvals, csums
            else:
                return poly[0], poly[1],  log(r_opt)
        else:
            return poly[0]


def corrdim_tangent_approx(x: np.ndarray, r_opt: float = None, norm_p=1, r_opt_ratio: float = 0.1, base: float = 10.0, full_output=False):
    if len(x.shape) == 1:
        x = x[:, np.newaxis]

    dim = x.shape[1]
    n_points = x.shape[0]

    if r_opt is None:
        r_opt = reference_rule(x, norm_p=norm_p)

    if base == 10:
        log = np.log10
    elif base == np.e:
        log = np.log
    else:
        raise AttributeError()

    r1 = r_opt
    r2 = r_opt * base**(-r_opt_ratio)
    # print(log(r1))
    # print(log(r2))
    c1 = corr_sum(x,  r1, norm_p=norm_p, allow_equals=False)
    c2 = corr_sum(x, r2, norm_p=norm_p, allow_equals=False)
    alpha = (log(c1) - log(c2)) / r_opt_ratio
    if full_output:
        r0 = log(r1)
        beta = log(c1) - alpha * r0

        return alpha, beta, r0
    else:
        return alpha


def approximate_k2(x: np.ndarray=None, r='auto', L_min=None, L_max=None, min_diag_number=0,
                   min_consecutive_nonzero_values=5, split_nonzero_slices=True, take_zero_splitted_slice='all', dt=1, method='avg', rp=None,
                   full_output=False, raise_if_empty=True, ordering=True):
# def approximate_k2(x: np.ndarray, r='auto', L_min=None, max_l=15, full_output=False, mrange: tuple=None, dt=1):
    assert(take_zero_splitted_slice in {'first', 'all'})
    assert(method in {'avg', 'fit'})

    try:
        from pyunicorn.timeseries import RecurrencePlot
    except ImportError:
        raise ImportError('The pyunicorn module is required to approximate K2. ')

    if r == 'auto':
        r = reference_rule(x, norm_p=float('inf'))
    # if max_l is None:
    #     max_l = -1

    if rp is None:
        if x is None:
            raise AttributeError("Either x or a pyunicorn RecurrencePlot must be supplied")
        rp = RecurrencePlot(x, threshold=r, silence_level=3, metric="supremum")

    if L_max is None:
        L_max = rp.max_diaglength()
    else:
        L_max = min(L_max, rp.max_diaglength())

    if L_min is not None:
        if isinstance(L_min, str):
            L_min = int(float(L_min.split()[0]) * rp.average_diaglength())

        L_min = min(L_min, 2)
    else:
        L_min = 2


    import itertools
    import operator
    # Entry i corresponds to diagline of size i
    diagline_dist = rp.diagline_dist()[L_min:L_max]
    diagline_dist_slices = []

    if split_nonzero_slices:
        nonzero_diagline_idx = [[i for i,value in it]
                                for key,it in itertools.groupby(enumerate(diagline_dist > min_diag_number), key=operator.itemgetter(1))
                                if key != 0]
        if take_zero_splitted_slice == 'all':
            diagline_dist_slices = [diagline_dist[s] for s in nonzero_diagline_idx
                                    if len(s) > min_consecutive_nonzero_values]
        elif take_zero_splitted_slice == "first":
            for s in nonzero_diagline_idx:
                if len(s) > min_consecutive_nonzero_values:
                    diagline_dist_slices = [diagline_dist[s]]
                    break

    else:
        diagline_dist_slices = [diagline_dist]



    if len(diagline_dist_slices) == 0:
        if raise_if_empty:
            raise ValueError('Cannot find sufficient support to estimate K2. Please check the timeseries for NaN or inf '
                             'values or manually provide a valid recurrence plot.')
        else:
            return np.nan

    k2_list = []

    if method == 'fit':
        x_vals = []
        y_vals = []
        for N_l in diagline_dist_slices:
            # bad_vals_filter = N_l > min_diags
            # bad_vals_filter[1:] = np.logical_and(bad_vals_filter[1:], N_l[:-1] - N_l[1:] != 0)
            # bad_vals_filter = N_l[:-1] - N_l[1:] != 0
            x_vals.append(np.arange(N_l.shape[0])[::-1])
            # x_vals = x_vals[bad_vals_filter]

            # N_l = N_l[bad_vals_filter]
            y_vals.append(np.log(N_l))
            # N_l = np.log(N_l)
        x_vals = np.concatenate(x_vals)
        y_vals = np.concatenate(y_vals)

        poly = lstsqr(_lstsqr_design_matrix(x_vals), y_vals)
        k2_list.append(poly[0])

    elif method == 'avg':
        # D_l = np.zeros((np.sum((len(s) for s in diagline_dist_slices)) - 1, ), dtype=float)

        for N_l in diagline_dist_slices:
            # D_l = np.zeros((N_l.shape[0]-1,), dtype=float)
            D_l = []
            for i in range(N_l.shape[0] - 1):
                if N_l[i + 1] != 0 and N_l[i] != 0:
                    if ordering:
                        # li = np.log(N_l[i])
                        # lip1 = np.log(N_l[i+1])

                        # D_l[i] = li - lip1
                        # D_l.append(li - lip1)
                        if  N_l[i+1] <= N_l[i]:
                            D_l.append(np.log(N_l[i] / N_l[i+1]))
                    else:
                        D_l.append(np.log(N_l[i] / N_l[i+1]))
            # D_l = np.nan_to_num(D_l)
            # D_l = D_l[D_l == D_l]

            k2_list.append(np.mean(D_l))

    if full_output:
        return [k2 / dt for k2 in k2_list]
    else:
        return np.mean(k2_list) / dt





from ..systems import ContinuousSys
from ..tools import nd_rand_init

import numpy as np


class Rossler(ContinuousSys):
    def __init__(self, a=0.1, b=0.1, c=14, n_points=2000, t_min=0, t_max=100):
        def ode(X, t):
            x,y,z = X
            return np.asarray([
                - y - z,
                x + a * y,
                b + z * (x - c)
            ])

        def rand_init():
            return nd_rand_init([-10,10],[-10,10],[0,25])

        super().__init__(dim=3, map_func=ode, init_func=rand_init, n_points=n_points, t_min=t_min, t_max=t_max)



import joblib
import numpy as np
import functools

from ..tools.io_conversion import _make_array


def nd_function(func):
    """
    Wraps a function to apply over slices of an n-dimensional array, specifying mandatory "axis" keyword.
    """
    @functools.wraps(func)
    def _wrapped_func(x, *args, **kwargs):
        if 'axis' in kwargs.keys(): 
            if not isinstance(kwargs['axis'], int):
                raise ValueError('Axis argument must have integer type.')

            axis = kwargs.pop('axis')
            if 'keepdims' in kwargs.keys(): 
                keepdims = kwargs.pop('keepdims')
            else: 
                keepdims = False


            _x = _make_array(x)

            _x = np.swapaxes(_x, axis, -1)
            _shape = _x.shape
            _x = _x.reshape((-1, _x.shape[-1]))

            results = joblib.Parallel(-1)(
                joblib.delayed(func)(_x[i], *args, **kwargs) 
                for i in range(_x.shape[0]))


            result_shape = np.shape(results[0])
            if len(result_shape) == 0:
                result_shape = (1,)

            results = np.asarray(results)

            if result_shape[0] == 1 and len(result_shape) == 1 and not keepdims: 
                results = results.reshape((*_shape[:-1],))
            else: 
                results = results.reshape((*_shape[:-1], *result_shape))
                results = np.swapaxes(results, axis, -1)

            return results

        else: 
            return func(x, *args, **kwargs)

    return _wrapped_func


def windowed_function(func):
    """
    Wraps a function to apply over windows, specified by mandatory keyword "window_size".
    """
    @functools.wraps(func)
    def _wrapped_func(x, *args, **kwargs):
        if 'window_size' in kwargs.keys(): 
            if not isinstance(kwargs['window_size'], int):
                raise ValueError('window size argument must have integer type.')
            window_size = kwargs.pop('window_size')

            if 'n_overlap' in kwargs.keys():

                if not isinstance(kwargs['n_overlap'], int):
                    raise ValueError('window size argument must have integer type.')
                n_overlap = kwargs.pop('n_overlap')
            else: 
                n_overlap = window_size // 2

            _x = _make_array(x)

            step_size = window_size - n_overlap
            n_windows = x.shape[0] // step_size

            results = [
                func(_x[i * step_size:min(i * step_size + window_size, x.shape[0]+1)], *args, **kwargs)
                for i in range(n_windows)
            ]

            results = np.asarray(results)
            return results

        else: 
            return func(x, *args, **kwargs)

    return _wrapped_func


from .BaseSys import AbstractBaseSys, ContinuousSys, DiscreteSys, SysWrapper
from .Lorenz import Lorenz
from .HindmarshRose import HindmarshRose
from .Rossler import Rossler
from .HenonMap import HenonMap
from .Ueda import Ueda
from .VanDerPol import VanDerPol
from .LogisticMap import LogisticMap
from .JansenRit import JansenRit



from ..tools.nd_utils import nd_function
from ..tools.math_utils import n_ball_volume
from typing import Union

from scipy.special import gamma
from scipy import stats

import numpy as np


def reference_rule_alpha(p: Union[float, int], d: int):
    if d == 1:
        alpha = (12. * np.sqrt(np.pi)) ** (1./5)
    else:
        alpha = (
            (4. * (2. * np.sqrt(np.pi))**d * (3. * gamma(1+(d+2.)/p) * gamma(1+1./p))**2)
            / ((d + 2.) * n_ball_volume(d, p) * (gamma(1+3./p) * gamma(1.+d/p)) ** 2)
        ) ** (1./(d+4))
    return alpha


@nd_function
def reference_rule(x: np.ndarray, dim: Union[int, str] = 'auto', norm_p: Union[int, float, str] = 2) -> float:
    n = x.shape[0]
    if dim == 'auto':
        d = x.shape[1] if len(x.shape) == 2 else 1
    elif isinstance(dim, int):
        d = dim
    else:
        raise ValueError(f'dim must be "auto" or int. Got: {type(dim)} ({dim})')

    std = np.sqrt(x.var(axis=0, ddof=1).mean())
    iqr = stats.iqr(x)
    scale = min(std, iqr/1.34)

    gamma_n = n ** (-1/(d+4))

    if norm_p in ['manhattan', 'euclidean', 'supremum']:
        norm_p = ['manhattan', 'euclidean'].index(norm_p) + 1 if norm_p != 'supremum' else float('inf')

    alpha_p_d = reference_rule_alpha(norm_p, d)
    return gamma_n * alpha_p_d * scale




from pykeos.tools import correlation_sum
from pykeos.tools.correlation_sum import _corr_sum_numpy_backend
import unittest
import numpy as np


def _make_ts(size=100):
    p = np.random.permutation(size)
    x = np.asarray([0] * int(size / 2) + [1] * int(size / 2))
    return x[p]


class TestCorrelationSum(unittest.TestCase):
    x = _make_ts()

    def test_correlation_sum(self):
        csum = correlation_sum(self.x, 0.5, allow_equals=True)
        self.assertAlmostEqual(csum, 0.5)

    def test_correlation_sum_numpy_backend(self):
        csum = _corr_sum_numpy_backend(self.x[:, None], 0.5, float('inf'))
        self.assertAlmostEqual(csum, 0.5)


if __name__ == '__main__':
    unittest.main()


import numpy as np
from scipy.special import gamma
import pandas as pd
from sklearn.metrics import mutual_info_score
from typing import Union


# def n_ball_volume(dim, norm):
#     if norm == 1:
#         return 2**dim/np.math.factorial(dim)
#     elif norm == 2:
#         return np.pi**(dim/2.) / (gamma(dim/2. + 1))
#     elif norm == float('inf') or norm == "inf":
#         return 2 ** dim


def n_ball_volume(dim, norm_p):
    if norm_p == 1:
        from scipy.special import factorial
        return 2.**dim / factorial(dim)
    elif norm_p == 2:
        return (np.sqrt(np.pi))**dim / gamma(dim/2 + 1)
    elif norm_p == float('inf') or norm_p == "inf":
        return 2 ** dim
    elif 0 < norm_p < float('inf'):
        return (2 * gamma(1 / norm_p + 1))**dim / gamma(dim/norm_p + 1)


def n_sphere_area(dim, norm):
    if norm == 1:
        return 2 ** dim / np.math.factorial(dim - 1)
    elif norm == 2:
        return 2 * np.pi ** ((dim + 1) / 2.) / (gamma((dim + 1) / 2.))
    elif norm == float('inf') or norm == "inf":
        return dim * 2 ** dim


def nd_rand_init(*tuples_lo_hi):
    return np.random.uniform(low=[t[0] for t in tuples_lo_hi], high=[t[1] for t in tuples_lo_hi])


def make_uniform_kernel(dim, norm_p):
    return lambda u: 1./n_ball_volume(dim, norm_p) if u <= 1 else 0.


def mutual_information(x, y, size_orig, n_bins="smooth") -> float:
    assert(x.size > 0 and y.size > 0)
    assert(n_bins in [None, 'smooth'] or isinstance(n_bins, int))
    if n_bins is None:
        n_bins = int(np.floor(np.sqrt(size_orig / 5)))

    elif n_bins is 'smooth':
        n_bins = int(np.floor(np.sqrt(size_orig / 3)))

    c_xy = np.histogram2d(x, y, n_bins)[0]
    return mutual_info_score(None, None, contingency=c_xy)


def lagged_mi(x, lag=1, *args, **kwargs) -> float:
    assert(lag >= 0)
    if lag > 0:
        return mutual_information(x[:-lag], x[lag:], x.shape[0], *args, **kwargs)
    else:
        return mutual_information(x, x, x.shape[0], *args, **kwargs)


def _lstsqr_design_matrix(x, degree=1):
    if degree > 1:
        raise NotImplementedError()
    return np.concatenate([x[:,None], np.ones((x.shape[0], 1))], axis=1)


def lstsqr(X, y):
    Xt = np.transpose(X)
    return np.linalg.pinv(Xt.dot(X)).dot(Xt).dot(y)


def sigmoid(x):
    return 1. / (1 + np.exp(-x))



from ..systems import ContinuousSys
from ..tools import nd_rand_init, sigmoid

import numpy as np



class JansenRit(ContinuousSys):
    def __init__(self, A=3.25, B=22, a_inv=10, b_inv=20, C=135, Crep=[1., 0.8, 0.25, 0.25], vmax=5, v0=6, r=0.56, n_points=5000, t_min=0, t_max=30):
        self.A = A
        self.B = B
        self.C = C
        self.a = 1. / a_inv
        self.b = 1. / b_inv
        self.C = C
        self.C1 = Crep[0] * C
        self.C2 = Crep[1] * C
        self.C3 = Crep[2] * C
        self.C4 = Crep[3] * C
        self.vmax = vmax
        self.v0 = v0
        self.r = r

        def sigm(x): 
            return self.vmax * sigmoid(self.r * (x - v0))


        def ode(X, t, p=None):
            x0, x1, x2, x3, x4, x5 = X
            p = p(t) if callable(p) else 0
            return np.asarray([
                x3, x4, x5, 
                self.A * self.a * sigm(x1 - x2) - 2 * self.a * x3 - self.a**2 * x0,
                self.A * self.a * (p + self.C2  * sigm(x1 * self.C1)) - 2 * self.a * x4 - self.a**2 * x1, 
                self.B * self.b * self.C4 * sigm(self.C3 * x0) - 2 * self.b * x5 - self.b ** 2 * x2
            ])

        def rand_init():
            return nd_rand_init(*[(-5,5)]*6)

        super().__init__(dim=6, map_func=ode, init_func=rand_init, n_points=n_points, t_min=t_min, t_max=t_max)




from abc import ABC, abstractmethod
from typing import Union, Dict, Callable, Optional
from numbers import Number
from scipy.integrate import odeint

import plotly.graph_objs as go
import numpy as np
import warnings


class AbstractBaseSys(ABC):
    """ Abstract class representing common attributes of pykeos systems.

    This class is not supposed to be instancied but is a common template for
    pykeos systems. Defines common functions that are usually applicable to
    either formally-defined discrete or continuous (sampled) systems, as well
    as to unformally-defined (wrapped) systems.

    Parameters
    ----------
    dim : int
        The dimension of the system
    map_func : Callable
        A callable map that will be integrated with the defined integration scheme. Will set the attribute time_map.
    init_func : Callable, optional
        The initialization function that must be called to get a random initial point to the system. Will set the
        attribute rand_init.
    n_points : int, optional
        The number of points.

    Attributes
    ----------
    dim : int
        The dimension of the system.
    n_points : int
        The number of points.
    states : np.ndarray
        An array of shape (n_points, dim) by convention. The first time states is called, it will integrate the time_map
        (if defined) from an initial_state (returned by rand_init()) for a duration of n_points.
    time_vec : Optional[np.ndarray]
        The time vector associated with the time series. Is usually None for discrete systems.
    time_map : Callable
        A callable map that will be integrated with the defined integration scheme. See DiscreteSys and ContinuousSys.
    rand_init : Optional[Callable]
        The initialization function that must be called to get a random initial point to the system.
    """
    def __init__(self, dim: int, map_func: Callable, init_func: Optional[Callable] = None,
                 n_points: Optional[int] = None):
        if dim < 1:
            raise ValueError("Incorrect number of dimensions")

        self.dim = dim
        self._states = None
        self.n_points = n_points
        self.time_vec = None

        self.rand_init = init_func if callable(init_func) else None
        self._initial_state = None

        if not callable(map_func):
            raise ValueError()
        self.time_map = map_func

    def __str__(self) -> str:
        return self.__class__.__name__

    @property
    def states(self) -> np.ndarray:
        """ Getter for the states timeseries of the system, calls integrate if the states are not yet defined.

        Returns
        -------
        np.ndarray
            An array of shape (n_points, dim) representing a timeseries of states for the system. If the
            serie is not defined yet, the system integrates its map function with its default parameters
            to create a timeserie.
        """
        if self._states is None:
            try:
                self.integrate()
            except ValueError:
                raise ValueError("Cannot integrate - call integrate manually with keyword update_states=True.")
        return self._states

    @states.setter
    def states(self, new_states: np.ndarray):
        """ Setter for the system's states.

        Parameters
        ----------
        new_states : np.ndarray
            A new timeseries of states of shape (new_n_points, dim)."""
        assert(len(new_states.shape) in [1, 2])
        if len(new_states.shape) == 2:
            assert(new_states.shape[1] == self.dim)
        self._states = new_states
        self.n_points = new_states.shape[0]

    @property
    def initial_state(self) -> Union[np.ndarray, float]:
        """ Getter for the initial state.

        Returns
        -------
        Union[np.ndarray, float]
            An array representing an initial point for the system. Will be used to solve the Initial
            Value Problem in the case of DiscreteSys and ContinuousSys.
        """
        if self._initial_state is None:
            if not callable(self.rand_init):
                raise ValueError("Either x0 or an initialization function must be provided")
            else:
                self._initial_state = self.rand_init()
        return self._initial_state

    @initial_state.setter
    def initial_state(self, x0: Union[np.ndarray, float]):
        """ Setter for the initial state.

        Parameters
        ----------
        x0 : Union[np.ndarray, float]
            An initial value that will be used to solve the Initial Value Problem in the case of DiscreteSys and
            ContinuousSys.
        """
        self._initial_state = x0

    @abstractmethod
    def integrate(self, *args, **kwargs) -> np.ndarray:
        """ The integration scheme to use in the case of formally-defined systems.

        This method is intended to do the all of the necessary integration work for typical types of systems.
        
        See also
        --------
        DiscreteSys.integrate
        ContinuousSys.integrate

        """
        pass

    def plot(self, show: bool = True, fig: Optional[go.Figure] = None, transient_index: int = 0, mode: str = "lines",
             fig_kwargs: Optional[Dict] = None, **trace_kwargs) -> go.Figure:
        """ Plots the states attribute using plotly for 1-, 2- or 3-dimensional systems.

        Parameters
        ----------
        show : bool
            Whether to show the figure or just return it.
        fig : Optional[go.Figure]
            An optional figure to draw on.
        mode : std
            A scatterplot mode compatible with plotly (e.g. 'lines', 'markers', 'lines+markers')
        transient_index : int
            Stop index for transients. We will omit [:transient_index].
        fig_kwargs : Optional[Dict]
            Kwargs to unpack to the fig.add_trace function.
        trace_kwargs :
            kwargs to unpack to the go.Scatter function.
        """
        if self.dim > 3:
            raise NotImplementedError()

        if fig is None:
            fig = go.Figure()

        if fig_kwargs is None:
            fig_kwargs = {}

        if transient_index is None:
            states = self.states
        else:
            states = self.states[transient_index:]

        if self.dim == 1:
            fig.add_trace(go.Scatter(x=self.time_vec if self.time_vec is not None else np.arange(self.n_points),
                                     y=states[:, 0], mode=mode, **trace_kwargs), **fig_kwargs)
        elif self.dim == 2:
            fig.add_trace(go.Scatter(x=states[:, 0], y=states[:, 1], mode=mode, **trace_kwargs),
                          **fig_kwargs)
        elif self.dim == 3:
            fig.add_trace(go.Scatter3d(x=states[:, 0], y=states[:, 1], z=states[:, 2], mode=mode,
                                       **trace_kwargs), **fig_kwargs)

        if show:
            fig.show()

        return fig

    def delay_coordinates(self, dim: Optional[int] = None, lag: Union[int, str] = "auto", axis: int = 0, *args, **kwargs):
        """ A convenient wrapper around pykeos delay_coordinates function.

        See also
        --------
        pykeos.tools.ebdg_utils.delay_coordinates
        """
        if dim is None:
            dim = self.dim
        from ..tools import delay_coordinates
        return delay_coordinates(self.states, dim=dim, lag=lag, axis=axis, *args, **kwargs)


class SysWrapper(AbstractBaseSys):
    """ A convenient wrapper to easily interface external data in pykeos framework.

    SysWrapper aims at wrapping a little bit of everything. It may represent partially
    observed systems, delay coordinates systems as well as real data from sensors. The usage is straightforward :
    any timeseries in form of a 1- or 2-dimensional array can be wrapped into a pykeos system using SysWrapper

    See also
    --------
    pykeos.tools.conv_utils : Contains helper function to convert data from and to pykeos.
    """
    def __init__(self, ts: np.ndarray, dim: Optional[int] = None, n_points: Optional[int] = None,
                 time_vec: Optional[np.ndarray] = None, t_min: Optional[Number] = None, t_max: Optional[Number] = None):
        """
        Parameters
        ----------
        ts : np.ndarray
            The
        :param ts:
        :param dim:
        :param n_points:
        :param time_vec:
        :param t_min:
        :param t_max:
        """
        if dim is None: 
            dim = ts.shape[1] if len(ts.shape) > 1 else 1
        if n_points is None: 
            n_points = ts.shape[0]
        if time_vec is None: 
            if t_min is None: 
                t_min = 0
            if t_max is None: 
                t_max = n_points
            time_vec = np.linspace(t_min, t_max, n_points)

        super(SysWrapper, self).__init__(
            dim=dim,
            map_func=lambda x: warnings.warn("The map function of a <SysWrapper> is not supposed to be called."),
            n_points=n_points
        )

        self.dim = dim
        self.n_points = n_points
        self.time_vec = time_vec
        self._states = ts if len(ts.shape) > 1 else ts[:, None]
        self._initial_state = self._states[0]

    def integrate(self, *args, **kwargs):
        raise RuntimeError("Instance of <SysWrapper> is not supposed to be integrated.")

    
class DiscreteSys(AbstractBaseSys):
    def integrate(self, n_points=None, x0=None, update_states=True):
        if x0 is not None:
            self.initial_state = x0

        if n_points is None:
            if self.n_points is None:
                raise ValueError("n_points must be provided")
            else:
                n_points = self.n_points

        states = [self.initial_state]

        for i in range(n_points - 1):
            states.append(self.time_map(states[i]))

        states = np.asarray(states)
        if update_states:
            self.states = states
            self.n_points = n_points
            return self

        else:
            return states


class ContinuousSys(AbstractBaseSys):
    def __init__(self, dim: int, map_func, init_func=None, n_points=None, t_min=None, t_max=None):
        super().__init__(dim, map_func, init_func, n_points=n_points)

        if None not in [n_points, t_min, t_max]:
            self.time_vec = np.linspace(t_min, t_max, n_points)
        else:
            self.time_vec = None

    def integrate(self, time_vec=None, x0=None, update_states=True, **odeint_kwargs):
        """ Solves an initial value problem with odeint.
        
        See Also
        --------
        DiscreteSys.integrate
        """
        if x0 is not None:
            self.initial_state = x0

        if time_vec is None:
            if self.time_vec is None:
                raise ValueError("time_vec must be provided")
            else:
                time_vec = self.time_vec

        states = odeint(self.time_map, self.initial_state, time_vec, **odeint_kwargs)

        if update_states:
            self.states = states
            self.time_vec = time_vec
            return self

        else:
            return states



from setuptools import find_packages, setup
from Cython.Build import cythonize

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='pykeos',
    version='0.0.3',
    author='Johan Medrano',
    python_requires='>=3.4',
    ext_modules=cythonize('pykeos/tools/_impl/impl.pyx'),
    author_email='',
    description='',
    long_description=long_description,
    platforms='Linux',
    packages=find_packages(),
    #package_dir={
    #    '': 'pykeos'
    #},
    include_package_data=True,
    install_requires=(
        'nolds',
        "plotly",
        'plotly_express'
    ),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



# http://thor.physics.wisc.edu/pubs/paper255.pdf
import numpy as np


from ..systems import ContinuousSys
from ..tools import nd_rand_init


class Ueda(ContinuousSys):
    def __init__(self, k=0.05, A=7.5, n_points=10000, t_min=0, t_max=100):
        def ode(X, t):
            _x, _y, _z = X
            return np.asarray([
                _y,
                - _x ** 3 - k * _y + A * np.sin(_z),
                1
            ])

        def rand_init():
            return nd_rand_init([-10, 10], [-10, 10], [-np.pi, np.pi])

        super().__init__(dim=3, map_func=ode, init_func=rand_init, n_points=n_points, t_min=t_min, t_max=t_max)



from ._impl.impl import _localized_diagline_histogram, \
    _localized_vertline_histogram, _localized_white_vertline_histogram, \
    _weighted_diagline_histogram, _weighted_vertline_histogram, _weighted_white_vertline_histogram

import numpy as np


def localized_diagline_histogram(recmat):
    n_times = int(recmat.shape[0])
    diagline_d = _localized_diagline_histogram(n_times, recmat)
    return diagline_d


def localized_vertline_histogram(recmat):
    n_times = int(recmat.shape[0])
    vertline_d = _localized_vertline_histogram(n_times, recmat)
    return vertline_d


def localized_white_vertline_histogram(recmat):
    n_times = int(recmat.shape[0])
    white_vertline_d = _localized_white_vertline_histogram(n_times, recmat)
    return white_vertline_d


def weighted_diagline_histogram(recmat, scale, order=2):
    recmat = recmat.astype(np.int8)
    n_times = recmat.shape[0]
    diag_hist = localized_diagline_histogram(recmat)
    max_diag_length = int(max(_[2] for _ in diag_hist) + 1)
    if max_diag_length > n_times:
        raise RuntimeError(f"max_diag_length ({max_diag_length}) is greater that n_times ({n_times})")
    weighted_histogram = np.zeros((n_times, max_diag_length), dtype=float)
    _weighted_diagline_histogram(n_times, diag_hist, weighted_histogram, scale, order)
    return weighted_histogram


def weighted_vertline_histogram(recmat, scale, order=2):
    recmat = recmat.astype(np.int8)
    n_times = recmat.shape[0]
    vert_hist = localized_vertline_histogram(recmat)
    max_vert_length = int(max(_[2] for _ in vert_hist) + 1)
    if max_vert_length > n_times:
        raise RuntimeError(f"max_vert_length ({max_vert_length}) is greater that n_times ({n_times})")
    weighted_histogram = np.zeros((n_times, max_vert_length), dtype=float)
    _weighted_vertline_histogram(n_times, vert_hist, weighted_histogram, scale, order)
    return weighted_histogram


def weighted_white_vertline_histogram(recmat, scale, order=2):
    recmat = recmat.astype(np.int8)
    n_times = recmat.shape[0]
    white_vert_hist = localized_white_vertline_histogram(recmat)
    max_white_vert_length = int(max(_[2] for _ in white_vert_hist) + 1)
    if max_white_vert_length > n_times:
        raise RuntimeError(f"max_white_vert_length ({max_white_vert_length}) is greater that n_times ({n_times})")
    weighted_histogram = np.zeros((n_times, max_white_vert_length), dtype=float)
    _weighted_white_vertline_histogram(n_times, white_vert_hist, weighted_histogram, scale, order)
    return weighted_histogram

